from appyx.layers.interface.presenters.base.presenter import Presenter


class StandardResponseContentPresenter(Presenter):

    def __init__(self, an_object, errors) -> None:
        super().__init__()
        self._an_object = an_object
        self._errors = errors

    def present(self):
        return {
            "object": self._an_object,
            "errors": self._errors
        }



from appyx.layers.domain.messages.error_messages import ErrorMessages


class ExceptionMessage(Exception):
    def __init__(self, message):
        self._message = message

    def __str__(self):
        return str(self._message)


class CreationError(ExceptionMessage):
    def __init__(self, class_been_instantiated, error_while_instantiating):
        message = ErrorMessages().creation_error_message(class_been_instantiated, error_while_instantiating)
        super(CreationError, self).__init__(message)


class ArgumentNotFound(ExceptionMessage):
    def __init__(self, argument_name):
        message = ErrorMessages().argument_not_found_message(argument_name)
        super(ArgumentNotFound, self).__init__(message)


class ParameterNotFoundException(ExceptionMessage):
    def __init__(self, missing_parameter_name):
        message = ErrorMessages().parameter_name_not_found(missing_parameter_name)
        super(ParameterNotFoundException, self).__init__(message)


class DuplicateParameterException(ExceptionMessage):
    def __init__(self, duplicated_parameter_name):
        message = ErrorMessages().duplicated_parameter_name(duplicated_parameter_name)
        super(DuplicateParameterException, self).__init__(message)


class TypeOfMessageParameterMissingExamplesException(ExceptionMessage):
    """
    This object models a lack of examples in the parameter list of a type of message.
    """
    def __init__(self, parameter_with_missing_example):
        message = ErrorMessages().parameter_example_missing(parameter_with_missing_example)
        super(TypeOfMessageParameterMissingExamplesException, self).__init__(message)



from unittest import TestCase


class AppyxTest:
    """
    This class acts as a wrapper for the abstract test classes below.
    This allows us to define abstract test methods that don't get executed by the testing framework.
    """
    class ProtoInteractionTest(TestCase):
        """
        This class provides tests for the behavior defined at the ProtoInteraction level.
        If you create a TestCase to test a ProtoInteraction subclass, you should inherit from this class
        to guarantee that the abstract behavior is still followed by the concrete class.
        """
        def test_a_proto_interaction_does_not_raise_an_exception_when_executed(self):
            # Given a sample proto interaction
            interaction = self._new_sample_proto_interaction()
            exception_not_raised = True

            # When we execute it
            try:
                interaction.execute()
            except:
                exception_not_raised = False

            # Then it does not raise an exception
            self.assertTrue(exception_not_raised, "Proto interactions should not raise exceptions during execution")

        def test_proto_interaction_answers_a_result_when_executed(self):
            # Given a sample proto interaction
            interaction = self._new_sample_proto_interaction()

            # When we execute it
            result = interaction.execute()

            # Then we get a result
            self.assertTrue(result.is_successful() or result.has_errors())

        def _new_sample_proto_interaction(self):
            raise NotImplementedError("Subclass responsibility")



from appyx.layers.application.interactions.parameters.set_of_parameters import SetOfParameters


class InteractionParameters(SetOfParameters):
    """
    I model the set of parameters and arguments that an interaction handles.
    """

    def _command(self):
        return self._interaction()

    def _interaction(self):
        return self._context



class Interaction:
    """
    This class models an interaction between the application and some interface.
    This class follows the Command pattern.
    Its responsible of executing the interaction and answering a Result.
    It encapsulates the logic of interacting with an external actor regardless of the type of interface.
    All handlers of external requests (command line, web, gui, etc) should collaborate with instances of this class.
    """
    def execute(self):
        raise NotImplementedError('Subclass responsibility')



from appyx.layers.application.application import Application
from appyx.layers.domain.messages.exceptions import DuplicateParameterException
from appyx.layers.application.interactions.app_interaction import AppInteraction
from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.application.interactions.proto_interaction_test import AppyxTest


class AppInteractionTest(AppyxTest.ProtoInteractionTest):
    """
    Tests the behavior of the abstract class AppInteraction (ex AppCommand)
    Note: These tests should be split in two:
        1 - Tests that validate appyx behavior
        2 - Tests that are inheritable and that are run on concrete subclasses
    """

    def _new_sample_proto_interaction(self):
        return self.app_command()

    def test_a_new_interaction_doesnt_have_parameters(self):
        # Given a new interaction
        cmd = self.app_command()

        # When we ask for its parameters
        parameters = cmd.get_parameters()

        # Then an empty list is returned
        self.assertEqual(parameters.size(), 0)

    def app(self):
        return Application()

    def app_command(self):
        return AppInteraction(app=self.app())

    def test_a_new_interaction_can_have_added_a_new_parameter(self):
        # Given a new interaction
        cmd = self.app_command()

        # When we add a new parameter
        parameter_price = Parameter(u"price", [])
        cmd.add_parameter(parameter_price)

        # Then we can retrieve it in various ways
        parameters = cmd.get_parameters()
        self.assertEqual(parameters.size(), 1)
        self.assertEqual(parameter_price, cmd.get_parameter_named(u"price"))

    def test_setting_parameters_erases_the_previous_ones(self):
        # Given a new interaction with three parameters
        cmd = self.app_command()
        cmd.add_parameters([Parameter(u"price", []), Parameter(u"currency", []), Parameter(u"is_cash", [])])
        self.assertEqual(cmd.get_parameters().size(), 3)

        # When we add a set parameters
        parameter_price = Parameter(u"price", [])
        cmd.set_parameters([parameter_price])

        # Then the amount of parameters is 1
        parameters = cmd.get_parameters()
        self.assertEqual(parameters.size(), 1)
        self.assertEqual(parameter_price, cmd.get_parameter_named(u"price"))

    def test_an_interaction_cannot_have_two_parameters_with_the_same_name(self):
        # Given an interaction with one parameter
        cmd = self.app_command()
        parameter_price = Parameter(u"price", [])
        cmd.add_parameter(parameter_price)

        # When we add a new parameter
        parameter_price = Parameter(u"price", [])
        exception = None
        try:
            cmd.add_parameter(parameter_price)
        except DuplicateParameterException as error:
            exception = error

        # Then an error occurred
        self.assertEqual("Parameter price is already present", str(exception))

    def test_an_interaction_cannot_have_a_parameter_named_none(self):
        # Given an interaction with one parameter
        cmd = self.app_command()

        # When we add a new parameter
        # noinspection PyTypeChecker
        parameter_precio = Parameter(None, [])
        exception = None
        try:
            cmd.add_parameter(parameter_precio)
        except ValueError as error:
            exception = error

        # Then an error occurred
        self.assertIsNotNone(exception)

    def test_an_interaction_cannot_have_a_parameter_named_empty_string(self):
        # Given an interaction with one parameter
        cmd = self.app_command()

        # When we add a new parameter
        parameter_precio = Parameter('', [])

        # Then an error occurred
        self.assertRaises(ValueError, cmd.add_parameter, parameter_precio)

    def test_an_interaction_cannot_have_a_parameter_named_empty_unicode_string(self):
        # Given an interaction with one parameter
        cmd = self.app_command()

        # When we add a new parameter
        parameter_precio = Parameter(u"", [])
        exception = None
        try:
            cmd.add_parameter(parameter_precio)
        except ValueError as error:
            exception = error

        # Then an error occurred
        self.assertIsNotNone(exception)

    def test_a_parameter_with_no_validators_is_valid(self):
        # Given an interaction with a parameter without validators
        cmd = self.app_command()
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)

        # When we ask can_execute
        result = cmd.validate()

        # Then the result is successful
        self.assertTrue(result.is_successful())

    def test_a_parameter_can_be_defined_as_required(self):
        # Given an interaction with a parameter
        cmd = self.app_command()
        parameter = Parameter(u"name", [])
        cmd.add_parameter(parameter)

        # When we set it as required
        cmd.add_required_parameter_named(u"name")

        # Then it's name appears in the required parameters list
        self.assertIn(u"name", cmd.required_parameter_names())

    def test_a_required_parameter_with_an_argument_is_valid(self):
        # Given an interaction with a required parameter
        cmd = self.app_command()
        parameter = Parameter(u"name", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"name")
        # and a corresponding argument
        cmd.set_arguments({u"name": "Eustaquio"})

        # When we ask can_execute
        result = cmd.validate()

        # Then the result is successful
        self.assertTrue(result.is_successful())

    def test_a_required_parameter_with_no_arguments_is_invalid(self):
        # Given an interaction with a parameter without validators
        cmd = self.app_command()
        parameter = Parameter(u"name", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"name")

        # When we ask can_execute
        result = cmd.validate()

        # Then the result is not successful
        self.assertTrue(result.has_errors())
        self.assertTrue(result.has_error_with_code("parameter_missing_error"))

    def test_a_required_parameter_can_be_made_optional(self):
        # Given an interaction with a required parameter and no argument for it
        cmd = self.app_command()
        parameter = Parameter(u"name", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"name")

        # When we make it optional
        cmd.remove_required_parameter_named(u"name")

        # Then the interaction is valid
        result = cmd.validate()
        self.assertTrue(result.is_successful())

    def test_a_parameter_can_be_declared_as_required_twice(self):
        # Given an interactionwith a required parameter
        cmd = self.app_command()
        parameter = Parameter(u"name", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"name")
        cmd.add_required_parameter_named(u"name")
        # and a corresponding argument
        cmd.set_arguments({u"name": "Eustaquio"})

        # When we ask can_execute
        result = cmd.validate()

        # Then the result is successful
        self.assertTrue(result.is_successful())

    def test_a_parameter_declared_twice_as_required_can_be_made_optional(self):
        # Given an interactionwith double required parameter and no argument for it
        cmd = self.app_command()
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")
        cmd.add_required_parameter_named(u"nombre")

        # When we make it optional
        cmd.remove_required_parameter_named(u"nombre")

        # Then the interaction is valid
        result = cmd.validate()
        self.assertTrue(result.is_successful())

    def test_an_interaction_can_have_one_of_three_parameters_required(self):
        # Given an interaction with three parameters
        cmd = self.app_command()
        parameter1 = Parameter(u"telefono", [])
        parameter2 = Parameter(u"mail", [])
        parameter3 = Parameter(u"whatsapp", [])
        cmd.add_parameter(parameter1)
        cmd.add_parameter(parameter2)
        cmd.add_parameter(parameter3)

        # When we set one as required and define one of them
        cmd.add_at_least_one_constraint([u"telefono", u"mail", u"whatsapp"])
        cmd.set_arguments({u"mail": 'test@probado.com'})

        # Then it should be ready to execute
        result = cmd.validate()
        self.assertTrue(result.is_successful())

    def test_an_interaction_must_have_one_of_three_parameters_required_defined(self):
        # Given an interaction with three parameters
        cmd = self.app_command()
        parameter1 = Parameter(u"telefono", [])
        parameter2 = Parameter(u"mail", [])
        parameter3 = Parameter(u"whatsapp", [])
        cmd.add_parameter(parameter1)
        cmd.add_parameter(parameter2)
        cmd.add_parameter(parameter3)

        # When we set one as required and don't define any of them
        cmd.add_at_least_one_constraint([u"telefono", u"mail", u"whatsapp"])

        # Then it should NOT be ready to execute
        result = cmd.validate()
        self.assertTrue(not result.is_successful())

    def test_interaction_parameters_can_be_set_and_read_individually(self):
        # Given an interaction with one parameter
        cmd = self.app_command()
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)

        # When we set the parameter argument
        cmd.set_argument_value(u"nombre", "Pepe Glamour")

        # Then we can retrieve it
        value = cmd.get_argument_named(u"nombre")
        self.assertEqual("Pepe Glamour", value)



from string import Formatter

from appyx.layers.domain.messages.exceptions import TypeOfMessageParameterMissingExamplesException, \
    CreationError, ParameterNotFoundException


class TypeOfMessage:
    """
    Notes:
        There are many local imports in this class to avoid circular imports, since it models a base element in the
        domain, and may be imported by some classes it imports.
    """

    def __init__(self, meaning, text_template, parameter_list=None) -> None:
        if parameter_list is None:
            parameter_list = []

        from appyx.layers.application.interactions.parameters.set_of_parameters import SetOfParameters
        parameter_names = [parameter.name() for parameter in parameter_list]
        self._parameters = SetOfParameters(context=self, initial_parameters=parameter_list,
                                           initial_required_parameter_names=parameter_names)
        self._meaning = meaning
        self._text_template = text_template

        self._validate_parameters_and_text_template()

    @classmethod
    def new_with_parametrized_text(cls, meaning, parametrized_text):
        text = cls._text_template_from_parametrized_text_list(parametrized_text)
        parameter_list = cls._extract_parameters_from_parametrized_text(parametrized_text)
        return cls(meaning, text, parameter_list)

    def __str__(self):
        return self.text_template()

    def means(self, meaning_to_compare_to):
        return self._meaning == meaning_to_compare_to

    def has_same_meaning_as(self, another_message):
        return another_message.means(self._meaning)

    def example_text(self):
        argument_examples = self._parameters.argument_examples()
        return self._text_template.format(**argument_examples)

    def parameters(self):
        return self._parameters

    def text_template(self):
        return self._text_template

    def validate_arguments(self, arguments):
        from appyx import Result
        result = Result()
        self._parameters.validate_arguments_existence(arguments, result)
        self._parameters.validate_argument_values(arguments, result)
        return result

    # --- private methods ---

    @classmethod
    def _text_template_from_parametrized_text_list(cls, parametrized_text):
        text_template = ""
        for item in parametrized_text:
            if isinstance(item, str):
                text_template = text_template + item
            else:
                text_template = text_template + "{" + item.name() + "}"
        return text_template

    @classmethod
    def _extract_parameters_from_parametrized_text(cls, parametrized_text):
        return [item for item in parametrized_text if not isinstance(item, str)]

    def _validate_parameters_and_text_template(self):
        self._assert_all_text_template_placeholders_have_a_matching_parameter()
        self._assert_all_parameters_have_an_example_argument()

    def _assert_all_parameters_have_an_example_argument(self):
        parameters_missing_examples = self._parameters.parameters_missing_examples()
        if len(parameters_missing_examples) > 0:
            parameter = parameters_missing_examples[0]
            error = TypeOfMessageParameterMissingExamplesException(parameter)
            raise CreationError(self.__class__, error)

    def _assert_all_text_template_placeholders_have_a_matching_parameter(self):
        try:
            self._parameters.assert_all_parameter_names_are_present(self._parameter_names())
        except ParameterNotFoundException as exception:
            raise CreationError(self.__class__, exception)

    def _parameter_names(self):
        return [field[1] for field in Formatter().parse(self._text_template) if field[1] is not None]



from unittest import TestCase

from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.application.interactions.validators.generic import IsString, NotEmptyString, IsAnID
from appyx.layers.domain.errors import BaseErrorFactory
from appyx.layers.domain.messages.exceptions import CreationError
from appyx.layers.domain.messages.type_of_message import TypeOfMessage


class TypeOfMessageTest(TestCase):
    def setUp(self):
        super().setUp()
        self._error_factory = BaseErrorFactory()

    def test_it_can_be_queried_for_its_meaning(self):
        # Given a type of message
        service_unavailable_text = "This service is unavailable."
        service_unavailable_meaning = "service_unavailable"
        type_of_message = TypeOfMessage(meaning=service_unavailable_meaning, text_template=service_unavailable_text)

        # When it is queried for its meaning
        has_expected_meaning = type_of_message.means(service_unavailable_meaning)

        # Then it returns its meaning
        self.assertTrue(has_expected_meaning)

    def test_two_can_mean_the_same_with_different_text(self):
        # Given two types of message with the same meaning and different text
        english_text = "This service is unavailable."
        spanish_text = "Este servicio no está disponible."
        meaning = 'service_unavailable'
        english_type_of_message = TypeOfMessage(meaning=meaning, text_template=english_text)
        spanish_type_of_message = TypeOfMessage(meaning=meaning, text_template=spanish_text)

        # When they are compared
        they_are_the_same_message = english_type_of_message.has_same_meaning_as(spanish_type_of_message)

        # Then they are equal
        self.assertTrue(they_are_the_same_message)

    def test_it_can_give_an_example_text(self):
        # Given a type of message
        type_of_message = TypeOfMessage(
            meaning='user_disconnected',
            text_template="The user {name} is disconnected.",
            parameter_list=[Parameter(name="name", example_argument="julian")]
        )

        # When it is queried for a text example
        example_text = type_of_message.example_text()

        # Then it returns a text presentation using example arguments for its parameters
        self.assertEqual("The user julian is disconnected.", example_text)

    def test_it_can_be_defined_with_a_list_of_text_and_parameters(self):
        # Given a type of message
        user_name = Parameter(name="name", example_argument="julian")
        type_of_message = TypeOfMessage.new_with_parametrized_text(
            meaning="user_disconnected",
            parametrized_text=["The user ", user_name, " is disconnected."]
        )

        # When it is queried for a text example
        example_text = type_of_message.example_text()

        # Then it returns a text presentation using example arguments for its parameters
        self.assertEqual("The user julian is disconnected.", example_text)

    def test_it_cannot_be_created_if_parameter_names_dont_match_text_placeholders(self):
        # Given a text with placeholders and a list of parameters with names that don't match the text's placeholders
        text = "There have been {amount} accidents in the workplace this year."
        parameter_list = [Parameter(name="age", example_argument=13)]
        raised_exception = None

        # When it is attempted to create a type of message with that text and those parameters
        try:
            TypeOfMessage(meaning="yearly_workplace_accidents", text_template=text, parameter_list=parameter_list)
        except CreationError as exception:
            raised_exception = exception

        # Then no instance is created and an error is raised informing the mismatch
        self.assertEqual("Cannot create instance of TypeOfMessage: Parameter named: amount not found",
                         str(raised_exception))

    def test_it_cannot_be_created_if_examples_are_missing_from_parameters(self):
        # Given a parametrized text with missing examples
        parametrized_text = ["There have been ", Parameter(name="amount"), "accidents in the workplace this year."]
        raised_exception = None

        # When it is attempted to create a type of message with that text
        try:
            TypeOfMessage.new_with_parametrized_text("yearly_workplace_accidents", parametrized_text)
        except CreationError as exception:
            raised_exception = exception

        # Then no instance is created and an error is raised informing the missing examples
        self.assertEqual(
            "Cannot create instance of TypeOfMessage: Parameter with name amount missing an example argument",
            str(raised_exception))

    def test_it_shows_its_parameters(self):
        # Given a type of message
        parameter = Parameter(name="full_name", example_argument="Julian Arnesino")
        type_of_message = TypeOfMessage(
            meaning="self_declared_pep",
            text_template="The user named {full_name} has declared themselves as politically exposed.",
            parameter_list=[parameter]
        )

        # When it is queried for its parameters
        message_parameters = type_of_message.parameters()

        # Then it returns a set with its parameters
        self.assertEqual(1, message_parameters.size())
        self.assertEqual(parameter, message_parameters.get_parameter_named(parameter.name()))

    def test_it_shows_its_text_template(self):
        # Given a type of message
        actual_text_template = "The user named {full_name} has declared themselves self as politically exposed."
        type_of_message = TypeOfMessage(
            meaning="self_declared_pep",
            text_template=actual_text_template,
            parameter_list=[Parameter(name="full_name", example_argument="Julian Arnesino")]
        )

        # When it is queried for its parameters
        obtained_text_template = type_of_message.text_template()

        # Then it returns a set with its parameters
        self.assertEqual(actual_text_template, obtained_text_template)

    def test_its_string_conversion_is_its_text_template(self):
        # Given a type of message
        actual_text_template = "The user named {full_name} has declared themselves self as politically exposed."
        type_of_message = TypeOfMessage(
            meaning="self_declared_pep",
            text_template=actual_text_template,
            parameter_list=[Parameter(name="full_name", example_argument="Julian Arnesino")]
        )

        # When it is queried for its parameters
        string_representation = str(type_of_message)

        # Then it returns a set with its parameters
        self.assertEqual(actual_text_template, string_representation)

    def test_it_can_validate_argument_existence(self):
        # Given a type of message with validators
        type_of_message = TypeOfMessage(
            "self_declared_pep", "The user named {full_name} has placed an order with id {id}.",
            parameter_list=[
                Parameter(name="full_name", example_argument="John Doe"),
                Parameter(name="id", example_argument="123")
            ]
        )

        # When it is asked to validate some missing arguments
        result = type_of_message.validate_arguments({"id": 123})

        # Then the result of the validation informs of the missing argument
        self.assertIn(self._error_factory.parameter_missing_error("El parámetro full_name es obligatorio"),
                      result.errors())

    def test_it_can_validate_argument_values(self):
        # Given a type of message with validators
        type_of_message = TypeOfMessage(
            "self_declared_pep", "The user named {full_name} has placed an order with id {id}.",
            parameter_list=[
                Parameter(name="full_name", validators=[IsString(), NotEmptyString()], example_argument="John Doe"),
                Parameter(name="id", validators=[IsAnID()], example_argument="123")
            ]
        )
        invalid_id = "0asd"

        # When it is asked to validate some invalid arguments
        result = type_of_message.validate_arguments({"full_name": "Julian Arnesino", "id": invalid_id})

        # Then the result of the validation informs of the invalid argument
        validator_error = self._error_factory.simple_error('Este campo debe ser un ID válido')
        expected_error = self._error_factory.parameter_validation_error("id", invalid_id, [validator_error])
        self.assertIn(expected_error, result.errors())



from appyx.layers.domain.result import Result


class ObjectValidator:
    """
    Models a validator for an object.
    Answers whether an object is valid or not.
    It can answer a simple boolean
    or a Result with more meaningful data about how and why an object is or isn't valid.

    validate is the method which answers a simple boolean.
    validate_with_result is a method that answers a Result explaining the validity.
    Both methods can be called with a Result as optional parameter
    Note: really? it makes sense for validate_with_result, but for validate?
    """

    def validate(self, an_object, result=None):
        """
        :type an_object: object
        :type result: Result
        :rtype: bool
        """
        result = self.validate_with_result(an_object, result)
        return result.is_successful()

    def validate_with_result(self, an_object, result=None):
        """
        :type an_object: object
        :type result: Result
        :rtype: Result
        """
        result = self._defined_result(result)

        return self._validate_with_result(an_object, result)

    def _defined_result(self, result=None):
        if result is None:
            result = Result()
        return result

    @classmethod
    def default_datetime_format(cls):
        """Return a format to represent the date and time in ISO 8601 format"""

        return "%Y-%m-%dT%H:%M:%S.%f"

    # --- private methods ---

    def _validate_with_result(self, an_object, result):
        raise NotImplementedError("Subclass responsibility")



class ProtoDomainOperation:
    """
    Operation:
    Models an operation that can be done in the domain which is important.
    Important is defined as being something that may be referenced
    (i.e. "adding a user" is something only administrators can do),
    composed by several different operations,
    ("send notification" is used in various sensitive operations)
    used in a command, etc.
    Provides an execute method to perform the operation and decouples instantiation from execution.
    ProtoOperation:
    Models an Operation that isn't fully mature and only the basic logic is implemented.
    Arguments for this operation should be defined through setters or the constructor of the class.
    """
    def execute(self):
        """
            Performs a domain operation.
        """
        raise NotImplementedError("Subclass responsibility")



from appyx.layers.application.interactions.interaction import Interaction
from appyx.layers.application.interactions.validators.parameters import NoneValidation
from appyx.layers.domain.result import Result


class SimpleInteraction(Interaction):
    """
    Models a simple and basic interaction.
    Defines a minimal protocol that any interaction should have.
    This allows to subclass and implement the bare minimum.

    It is expected that subclasses will override at least _execute_from_successful_result.
    """
    def __init__(self, parameters=None):
        super().__init__()
        self._parameters = parameters or {}

    def set_parameters(self, parameters):
        self._parameters = parameters

    def validate(self):
        """
        Answers a result indicating the validity of the receiver.

        If result has no errors, it means the receiver can be executed (although it may fail if further problems arise).
        If result has errors, then the receiver can not be executed (and the errors explain why)
        """
        result = Result()
        for validator in self._parameters_validator():
            result = validator.validate(self._parameters, result)
        return result

    def execute(self):
        """
        Performs the operation of the receiver and answers the result of the execution.

        :return: result of execution
        """
        result = self.validate()
        if result.has_errors():
            return result
        return self._execute_from_successful_result(result)

    # --- private methods ---

    def _execute_from_successful_result(self, result):
        """
        Execute the command assuming can_execute is successful.
        My subclasses should override me.
        Performs the operation and answers the result of the execution

        :return: result of execution
        """
        return result

    def _parameters_validator(self):
        """
        My subclass should override me. Configures how my parameters must be validated.

        :return: a list of parameter names
        """
        return [NoneValidation()]



from unittest import TestCase

from appyx.layers.application.application import Application
from appyx.layers.application.interactions.app_interaction import AppInteraction
from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.domain.messages.exceptions import ArgumentNotFound, DuplicateParameterException


class TestApp(Application):
    pass


class CommandParametersTest(TestCase):
    def setUp(self) -> None:
        self._app = TestApp()

    def test_a_new_command_doesnt_have_parameters(self):
        # Given a new command
        cmd = AppInteraction(self._app)

        # When we ask for its parameters
        parameters = cmd.get_parameters()

        # Then an empty list is returned
        self.assertEqual(parameters.size(), 0)

    def test_a_new_command_can_have_added_a_new_parameter(self):
        # Given a new command
        cmd = AppInteraction(self._app)

        # When we add a new parameter
        parameter_precio = Parameter(u"precio", [])
        cmd.add_parameter(parameter_precio)

        # Then we can retrieve it in various ways
        parameters = cmd.get_parameters()
        self.assertEqual(parameters.size(), 1)
        self.assertEqual(parameter_precio, cmd.get_parameter_named(u"precio"))

    def test_setting_parameters_erases_the_previous_ones(self):
        # Given a new command with three parameters
        cmd = AppInteraction(self._app)
        cmd.add_parameters([Parameter(u"precio", []), Parameter(u"moneda", []), Parameter(u"es_contado", [])])
        self.assertEqual(cmd.get_parameters().size(), 3)

        # When we add a set parameters
        parameter_price = Parameter(u"price", [])
        cmd.set_parameters([parameter_price])

        # Then the amount of parameters is 1
        parameters = cmd.get_parameters()
        self.assertEqual(parameters.size(), 1)
        self.assertEqual(parameter_price, cmd.get_parameter_named(u"price"))

    def test_a_command_cannot_have_two_parameters_with_the_same_name(self):
        # Given a command with one parameter
        cmd = AppInteraction(self._app)
        parameter_precio = Parameter(u"precio", [])
        cmd.add_parameter(parameter_precio)

        # When we add a new parameter
        parameter_precio = Parameter(u"precio", [])
        exception = None
        try:
            cmd.add_parameter(parameter_precio)
        except DuplicateParameterException as error:
            exception = error

        # Then an error occurred
        self.assertEqual("Parameter precio is already present", str(exception))

    def test_a_command_cannot_have_a_parameter_named_none(self):
        # Given a command with one parameter
        cmd = AppInteraction(self._app)

        # When we add a new parameter
        # noinspection PyTypeChecker
        parameter_precio = Parameter(None, [])
        exception = None
        try:
            cmd.add_parameter(parameter_precio)
        except ValueError as error:
            exception = error

        # Then an error occurred
        self.assertIsNotNone(exception)

    def test_a_command_cannot_have_a_parameter_named_empty_string(self):
        # Given a command with one parameter
        cmd = AppInteraction(self._app)

        # When we add a new parameter
        parameter_precio = Parameter('', [])

        # Then an error occurred
        self.assertRaises(ValueError, cmd.add_parameter, parameter_precio)

    def test_a_command_cannot_have_a_parameter_named_empty_unicode_string(self):
        # Given a command with one parameter
        cmd = AppInteraction(self._app)

        # When we add a new parameter
        parameter_precio = Parameter(u"", [])
        exception = None
        try:
            cmd.add_parameter(parameter_precio)
        except ValueError as error:
            exception = error

        # Then an error occurred
        self.assertIsNotNone(exception)

    def test_a_parameter_with_no_validators_is_valid(self):
        # Given a command with a parameter without validators
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)

        # When we ask can_execute
        result = cmd.can_execute()

        # Then the result is successful
        self.assertTrue(result.is_successful())

    def test_a_parameter_can_be_defined_as_required(self):
        # Given a command with a parameter
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)

        # When we set it as required
        cmd.add_required_parameter_named(u"nombre")

        # Then it's name appears in the required parameters list
        self.assertIn(u"nombre", cmd.required_parameter_names())

    def test_a_required_parameter_with_an_argument_is_valid(self):
        # Given a command with a required parameter
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")
        # and a corresponding argument
        cmd.set_arguments({u"nombre": "Eustaquio"})

        # When we ask can_execute
        result = cmd.can_execute()

        # Then the result is successful
        self.assertTrue(result.is_successful())

    def test_a_required_parameter_with_no_arguments_is_invalid(self):
        # Given a command with a parameter without validators
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")

        # When we ask can_execute
        result = cmd.can_execute()

        # Then the result is not successful
        self.assertTrue(result.has_errors())
        self.assertEqual(len(result.errors()), 1)
        self.assertEqual(result.errors()[0].code(), "parameter_missing_error")

    def test_a_required_parameter_can_be_made_optional(self):
        # Given a command with a required parameter and no argument for it
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")

        # When we make it optional
        cmd.remove_required_parameter_named(u"nombre")

        # Then the command can execute
        result = cmd.can_execute()
        self.assertTrue(result.is_successful())

    def test_a_parameter_can_be_declared_as_required_twice(self):
        # Given a command with a required parameter
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")
        cmd.add_required_parameter_named(u"nombre")
        # and a corresponding argument
        cmd.set_arguments({u"nombre": "Eustaquio"})

        # When we ask can_execute
        result = cmd.can_execute()

        # Then the result is successful
        self.assertTrue(result.is_successful())

    def test_a_parameter_declared_twice_as_required_can_be_made_optional(self):
        # Given a command with double required parameter and no argument for it
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")
        cmd.add_required_parameter_named(u"nombre")

        # When we make it optional
        cmd.remove_required_parameter_named(u"nombre")

        # Then the command can execute
        result = cmd.can_execute()
        self.assertTrue(result.is_successful())

    def test_a_command_can_have_one_of_three_parameters_required(self):
        # Given a command with three parameters
        cmd = AppInteraction(self._app)
        parameter1 = Parameter(u"phone", [])
        parameter2 = Parameter(u"mail", [])
        parameter3 = Parameter(u"whatsapp", [])
        cmd.add_parameter(parameter1)
        cmd.add_parameter(parameter2)
        cmd.add_parameter(parameter3)

        # When we set one as required and define one of them
        cmd.add_at_least_one_constraint([u"phone", u"mail", u"whatsapp"])
        cmd.set_arguments({u"mail": 'test@probado.com'})

        # Then it should be ready to execute
        result = cmd.can_execute()
        self.assertTrue(result.is_successful())

    def test_a_command_must_have_one_of_three_parameters_required_defined(self):
        # Given a command with three parameters
        cmd = AppInteraction(self._app)
        parameter1 = Parameter(u"phone", [])
        parameter2 = Parameter(u"mail", [])
        parameter3 = Parameter(u"whatsapp", [])
        cmd.add_parameter(parameter1)
        cmd.add_parameter(parameter2)
        cmd.add_parameter(parameter3)

        # When we set one as required and don't define any of them
        cmd.add_at_least_one_constraint([u"phone", u"mail", u"whatsapp"])

        # Then it should NOT be ready to execute
        result = cmd.can_execute()
        self.assertTrue(not result.is_successful())

    def test_command_parameters_can_be_set_and_read_individually(self):
        # Given a command with one parameter
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)

        # When we set the parameter argument
        cmd.set_argument_value(u"nombre", "Pepe Glamour")

        # Then we can retrieve it
        value = cmd.get_argument_named(u"nombre")
        self.assertEqual("Pepe Glamour", value)

    def test_when_asked_for_a_missing_parameter_None_is_returned(self):
        # Given an interaction with a declared parameter
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")
        cmd.set_arguments({'nombre': 'Julian'})

        # When we ask for a parameter that is not declared
        parameter_asked = cmd.get_parameter_named("apellido")

        # Then a particular exception is raised
        self.assertIsNone(parameter_asked)

    def test_when_asked_for_a_missing_argument_an_exception_is_raised(self):
        # Given an interaction with a declared parameter
        cmd = AppInteraction(self._app)
        parameter = Parameter(u"nombre", [])
        cmd.add_parameter(parameter)
        cmd.add_required_parameter_named(u"nombre")
        cmd.set_arguments({'nombre': 'Julian'})
        exception_raised = None

        # When we ask for an argument that has no parameter
        try:
            cmd.get_argument_named("apellido")
        except ArgumentNotFound as error:
            exception_raised = error

        # Then a particular exception is raised
        self.assertEqual('Argument named: apellido not found', str(exception_raised))



import unittest

from appyx.layers.application.interactions.parameters.parameter import Parameter
from appyx.layers.application.interactions.validators.generic import NotNone, MinimumValue, IsADateString


class ParameterTest(unittest.TestCase):
    def test_a_parameter_with_no_validators_is_valid(self):
        parameter = Parameter('nombre', [])

        result = parameter.validate('Pedro')

        self.assertTrue(result.is_successful())
        self.assertEqual('nombre', parameter.name())

    def test_a_parameter_with_validators_and_invalid_argument_is_not_valid(self):
        parameter = Parameter('edad', [NotNone(), MinimumValue(0)])

        result = parameter.validate(-1)

        self.assertTrue(result.has_errors())

    def test_a_parameter_can_have_a_description(self):
        parameter = Parameter('inicio', [NotNone(), IsADateString('2021-08-04'), 'Fecha de inicio del trámite'])

        description = parameter.description()

        self.assertTrue('Fecha de inicio del trámite', description)



class AppyxHttpResponse:
    def __init__(self, http_status_code, description, content):
        self._http_status_code = http_status_code
        self._description = description
        self._content = content

    def http_status_code(self):
        return self._http_status_code

    def description(self):
        return self._description

    def content(self):
        return self._content



from appyx.layers.application.exception_handler import AddExceptionToResult


class Application:
    """
    Models an application.
    It holds resources and interactions that allow the logic and business rule to operate.

    An application defines external resources, how to handle the exceptions those might rise for each interaction.
    In the future it may hold workflows, which in turn would hold interactions.
    Interactions should obtain the resources they need from this object. Maybe passing them onto the domain.
    """

    def __init__(self, exception_handler=None) -> None:
        super().__init__()
        self._general_exception_handler = exception_handler or self._default_exception_handler()

    def general_exception_handler(self):
        return self._general_exception_handler

    def set_general_exception_handler(self, exception_handler):
        self._general_exception_handler = exception_handler

    def _default_exception_handler(self):
        return AddExceptionToResult()

    def current_business(self):
        from appyx.polls.domain.poll_station import PollStation, FakeClock, QuestionsArchive
        clock = FakeClock()
        business = PollStation(clock, QuestionsArchive())
        return business



from appyx.layers.domain.errors import BaseErrorFactory


class ExceptionHandler:
    """
    Models the policy on how to handle exceptions.
    """
    def handle(self, an_exception, result):
        raise NotImplementedError('Subclass responsibility')


class AddExceptionToResult(ExceptionHandler):
    """
    Models the policy of packaging an exception as an error within a result.
    """
    def handle(self, an_exception, result):
        result_error = BaseErrorFactory().unexpected_error(an_exception)
        result.add_error(result_error)
        return result


class NullExceptionHandler(ExceptionHandler):
    """
    Models the policy of not handling exceptions. Effectively letting them run through.
    """
    def handle(self, an_exception, result):
        raise an_exception



from unittest import TestCase

from appyx import Result
from appyx.layers.domain.errors import BaseErrorFactory


class ResultTest(TestCase):

    def test_a_result_with_no_errors_is_successful(self):
        result = Result()

        self.assertTrue(result.is_successful())

    def test_a_new_result_has_no_errors(self):
        result = Result()

        self.assertEqual(0, len(result.errors()))
        self.assertFalse(result.has_errors())

    def test_a_result_with_errors_is_not_successful(self):
        result = Result()
        result.add_error("A sample message")

        self.assertFalse(result.is_successful())
        self.assertEqual(1, len(result.errors()))
        self.assertTrue(result.has_errors())

    def test_a_result_with_a_string_error_has_an_error_code(self):
        result = Result()
        expected_error_code = 'simple_error_code'
        result.add_error("A sample message")

        error_code = result.error_codes()[0]

        self.assertEqual(expected_error_code, error_code)
        self.assertTrue(result.has_error_with_code(expected_error_code))
        self.assertTrue(result.has_at_least_one_error_with_codes([expected_error_code]))
        self.assertTrue(result.has_at_least_one_error_with_codes([expected_error_code, 'an_unexpected_error_code']))
        self.assertFalse(result.has_at_least_one_error_with_codes(['an_unexpected_error_code']))

    def test_a_result_can_remove_identical_errors(self):
        result = Result()
        repeated_error = BaseErrorFactory().simple_error('error error!')
        result.add_error(repeated_error)
        result.add_error(repeated_error)

        result.delete_repeated_errors()

        self.assertEqual(1, len(result.errors()))

    def test_a_result_can_remove_equal_errors(self):
        result = Result()
        result.add_error(BaseErrorFactory().simple_error('A simple error'))
        result.add_error(BaseErrorFactory().simple_error('A simple error'))

        result.delete_repeated_errors()

        self.assertEqual(1, len(result.errors()))



from unittest import TestCase

from appyx.layers.application.interactions.simple_interaction import SimpleInteraction


class InteractionTest(TestCase):

    def test_the_execution_of_an_interaction_answers_a_result(self):
        # Given an interaction
        interaction = SimpleInteraction()

        # When we execute it
        result = interaction.execute()

        # Then we get a result
        self.assertTrue(result.is_successful() or result.has_errors())

    def test_before_executing_it_we_can_validate_an_interaction(self):
        # Given an interaction
        interaction = SimpleInteraction()

        # When we validate it
        result = interaction.validate()

        # Then we get a result
        self.assertTrue(result.is_successful() or result.has_errors())

    def test_if_the_interaction_isnt_valid_the_execution_will_fail(self):
        # Given an invalid interaction
        interaction = self._invalid_interaction()
        result = interaction.validate()
        self.assertFalse(result.is_successful())

        # When we know it's invalid
        execution_result = interaction.execute()

        # Then the execution fails
        self.assertFalse(execution_result.is_successful())

    def _invalid_interaction(self):
        class InvalidInteraction(SimpleInteraction):
            def validate(self):
                result = super().validate()
                result.add_error("I am invalid")
                return result

        return InvalidInteraction()



class HTTPInterfaceSection:
    def __init__(self, endpoints, name) -> None:
        self._endpoints = endpoints
        self._name = name

    def endpoints(self):
        return self._endpoints

    def name(self):
        return self._name



from appyx.layers.domain.validators.object_validator import ObjectValidator


class ListValidator(ObjectValidator):
    '''
    List validator that currently implements two kind of validations:
        - Validations to the list as a whole
        - Validations that apply to a single element of the list
    '''
    def __init__(self, list_validators=None, element_validators=None):
        self._list_validators = list_validators or []
        self._element_validators = element_validators or []

    def _validate_with_result(self, list_to_validate, result):
        # Run validators that apply to the list as a whole
        for validator in self._list_validators:
            validator.validate_with_result(list_to_validate, result)

        # Run validators that apply to each element of the list
        for validator in self._element_validators:
            self._validate_for_each_element_of_list(list_to_validate, validator, result)

        return result

    def _validate_for_each_element_of_list(self, list_to_validate, validator, result):
        for element in list_to_validate:
            validator.validate_with_result(element, result)
        return result


from setuptools import setup, find_packages

setup(
    name='appyx',
    version='0.0.10',
    description='...',
    url='https://gitlab.com/eryx/appyx',
    author='Eryx Team',
    author_email='info@eryx.co',
    license='MIT',
    packages=find_packages(),
    install_requires=[],
)



from appyx.layers.application.interactions.validators.parameters import AtLeastOneParameterRequired, \
    AtLeastOneParameterNotNoneRequired
from appyx.layers.domain.errors import BaseErrorFactory
from appyx.layers.domain.messages.exceptions import ArgumentNotFound, ParameterNotFoundException, \
    DuplicateParameterException


class SetOfParameters:
    """
    I model the set of parameters and arguments that a context handles.
    I handle the definition and validation of the arguments based on the validations each parameter has defined.

    _command: context that holds the parameters
    _parameters: list of parameters that compose me
    _arguments = dictionary of names and values for the parameters I have
    _required_parameter_names: list of names of parameters that should be present in arguments
    _parameters_validators = list of validators that apply to all parameters and arguments as a whole

    Note:
        Parameter is the "slot" that I have for a value. It has a name and validators.
        Argument is the value for that parameter.
    """

    def __init__(self, context, initial_parameters=None,
                 initial_required_parameter_names=None, initial_at_least_one_required_sets=None,
                 default_arguments=None):
        _initial_parameters = initial_parameters or []
        _initial_required_parameter_names = initial_required_parameter_names or set()
        _initial_at_least_one_required_sets = initial_at_least_one_required_sets or []
        self._context = context
        self._parameters = _initial_parameters
        self._arguments = default_arguments or {}
        self._required_parameter_names = _initial_required_parameter_names
        self._parameters_validators = []
        for set_of_names in _initial_at_least_one_required_sets:
            self.add_at_least_one_constraint(set_of_names)

    def set_arguments(self, arguments):
        """Set the values that each parameter will have.
        Arguments is a dictionary with the names of the parameters as keys and the value of the argument as value.

        :param arguments: dict of str: object
        """
        self._arguments = arguments

    def set_argument_value(self, argument_name, value):
        """Set the value for a specific argument.

        :param argument_name: str
        :param value: object
        """
        self._arguments[argument_name] = value

    def add_parameter(self, parameter):
        if parameter.name() is None:
            raise ValueError('Parameter cannot be None')
        if parameter.name() == '':
            raise ValueError('Parameter cannot be an empty string')
        is_parameter_already_present = self.get_parameter_named(parameter.name()) is not None
        if is_parameter_already_present:
            raise DuplicateParameterException(parameter.name())
        self._parameters.append(parameter)

    def get_parameter_named(self, name):
        """Answers the parameter with name <name>.
        Answers None if no parameter with that named is found.

        :param name: str
        :return: Parameter or none
        """
        parameter = None
        for param in self._parameters:
            if param.name() == name:
                parameter = param
        return parameter

    def add_required_parameter_named(self, parameter_name):
        """Define parameter named <parameter_name> as required.

        :param parameter_name: str
        """
        self._required_parameter_names.add(parameter_name)

    def add_required_parameters_named(self, parameter_name_list):
        """Define parameters named [<parameter_name>] as required.

        :param parameter_name_list: List<str>
        """
        for parameter_name in parameter_name_list:
            self.add_required_parameter_named(parameter_name)

    def remove_required_parameter_named(self, parameter_name):
        """Remove the required restriction for parameter named <parameter_name>.

        :param parameter_name: str
        """
        self._required_parameter_names.remove(parameter_name)

    def required_parameter_names(self):
        return self._required_parameter_names

    def add_at_least_one_constraint(self, parameter_names):
        """Add a restriction that at least on of parameter_names should be required.

        :param parameter_names: list of str
        """
        self.assert_all_parameter_names_are_present(parameter_names)
        self.add_parameters_validator(AtLeastOneParameterRequired(parameter_names))

    def add_at_least_one_not_none_constraint(self, parameter_names):
        """Add a restriction that at least on of parameter_names should be required and not none.

        :param parameter_names: list of str
        """
        self.assert_all_parameter_names_are_present(parameter_names)
        self.add_parameters_validator(AtLeastOneParameterNotNoneRequired(parameter_names))

    def add_parameters_validator(self, validator):
        """Add a validator that analysis the parameters and arguments as a whole"""
        self._parameters_validators.append(validator)

    def validate_parameters_and_arguments(self, result):
        """
        Run validations of all arguments against all parameters and answer the result of such analysis.

        :param result: Result
        :return: Result
        """
        self._validate_unexpected_arguments(result)
        self._validate_missing_arguments(result)
        self._validate_present_arguments(result)

        if result.is_successful():
            self._apply_validators_for_parameters_set(result)

        return result

    def is_defined_argument_named(self, argument_name):
        return argument_name in self._arguments

    def get_argument_named(self, argument_name):
        if self.is_defined_argument_named(argument_name):
            return self._arguments[argument_name]
        else:
            raise ArgumentNotFound(argument_name)

    def size(self):
        """Answer the amount of parameters the receiver has.

        :return: int
        """
        return len(self._parameters)

    def clear_parameters(self):
        self._parameters = []

    def items(self):
        return self._arguments.items()

    def parameters_missing_examples(self):
        return [parameter for parameter in self._parameters if not parameter.has_example_argument()]

    def argument_examples(self):
        argument_examples = {}
        for parameter in self._parameters:
            argument_examples[parameter.name()] = parameter.example_argument()
        return argument_examples

    def assert_all_parameter_names_are_present(self, parameter_names):
        """Note: We delegate the raise of the exception to _parameter_named"""
        for name in parameter_names:
            self._parameter_named(name)

    def validate_arguments_existence(self, arguments, result):
        parameters_names = self._parameters_names()
        arguments_names = arguments.keys()
        missing_arguments = list(set(parameters_names) - set(arguments_names))
        self._handle_missing_arguments(missing_arguments, result)

    def validate_argument_values(self, arguments, result):
        for parameter in self._parameters:
            parameter_name = parameter.name()
            parameter_is_present = parameter_name in arguments.keys()
            if parameter_is_present:
                value = arguments[parameter_name]
                parameter.validate(value, result)

    # --- private methods ---

    def _validate_unexpected_arguments(self, result):
        parameters_names = self._parameters_names()
        arguments_names = self._arguments.keys()
        unexpected_arguments = list(set(arguments_names) - set(parameters_names))
        there_are_unexpected_parameters = len(unexpected_arguments) > 0
        if there_are_unexpected_parameters:
            # Do something with the extra parameters
            self._handle_unexpected_arguments(unexpected_arguments, result)

    def _handle_unexpected_arguments(self, extra_arguments, result):
        self._context.handle_unexpected_arguments(extra_arguments, result)

    def _validate_missing_arguments(self, result):
        self.validate_arguments_existence(self._arguments, result)

    def _handle_missing_arguments(self, missing_arguments, result):
        for argument in missing_arguments:
            parameter = self._parameter_named(argument)
            if parameter.name() in self.required_parameter_names():
                result.add_error(
                    BaseErrorFactory().parameter_missing_error(u"El parámetro {0} es obligatorio".format(argument)))

    def _parameter_named(self, argument):
        for parameter in self._parameters:
            if parameter.name() == argument:
                return parameter
        raise ParameterNotFoundException(missing_parameter_name=argument)

    def _parameters_names(self):
        names = [parameter.name() for parameter in self._parameters]
        return names

    def _validate_present_arguments(self, result):
        self.validate_argument_values(self._arguments, result)

    def _apply_validators_for_parameters_set(self, result):
        for validator in self._parameters_validators:
            validator.validate(self._arguments, result)



from unittest import TestCase

from appyx.layers.application.environment import Environment


class PollCreation(TestCase):
    def test_01_our_business_can_show_the_latest_polls(self):
        business = self._new_poll_station()

        latest_questions = business.latest_questions()

        self.assertEqual(0, len(latest_questions))

    def test_02_questions_can_be_created_and_published(self):
        business = self._new_poll_station()

        business.create_question_with_choices("What's up", ["Not much", "The sky", "Just hacking again"])

        latest_questions = business.latest_questions()
        self.assertEqual(1, len(latest_questions))

    def test_03_questions_have_the_correct_text(self):
        business = self._new_poll_station()

        question_text = "What's up"
        business.create_question_with_choices(question_text, ["Not much", "The sky", "Just hacking again"])

        latest_question = business.latest_questions()[0]
        self.assertEqual(question_text, latest_question.text())

    def test_04_questions_have_the_correct_creation_time(self):
        business = self._new_poll_station()
        clock = business.clock()
        question_creation_time = clock.current_time()

        business.create_question_with_choices("What's up", ["Not much", "The sky", "Just hacking again"])
        clock.advance_1_day()

        latest_question = business.latest_questions()[0]
        self.assertEqual(question_creation_time, latest_question.creation_time())

    def test_05_questions_have_the_correct_choices(self):
        business = self._new_poll_station()

        choices_texts = ["Not much", "The sky", "Just hacking again"]
        business.create_question_with_choices("What's up", choices_texts)
        latest_question = business.latest_questions()[0]
        self.assertEqual(choices_texts, latest_question.choices_texts())

    def test_06_a_newly_created_question_has_0_votes_on_its_choices(self):
        business = self._new_poll_station()
        business.create_question_with_choices("What's up", ["Not much", "The sky", "Just hacking again"])
        latest_question = business.latest_questions()[0]

        self.assertEqual(0, latest_question.choices()[0].amount_of_votes())
        self.assertEqual(0, latest_question.choices()[1].amount_of_votes())
        self.assertEqual(0, latest_question.choices()[2].amount_of_votes())

    def test_07_a_vote_can_be_casted_on_a_question_choice(self):
        business = self._new_poll_station()
        business.create_question_with_choices("What's up", ["Not much", "The sky", "Just hacking again"])
        latest_question = business.latest_questions()[0]

        business.vote(latest_question, latest_question.choices()[1])

        latest_question = business.latest_questions()[0]
        self.assertEqual(0, latest_question.choices()[0].amount_of_votes())
        self.assertEqual(1, latest_question.choices()[1].amount_of_votes())
        self.assertEqual(0, latest_question.choices()[2].amount_of_votes())

    # --- private methods ---

    def _new_poll_station(self):
        business = Environment().current_app().current_business()
        return business



from appyx.layers.application.interactions.interaction import Interaction


class ProtoInteraction(Interaction):
    """
    Models an interaction that is not yet fully designed.
    It encapsulates and isolates the logic in one method: execute.
    The execute method must return an instance of Result.
    The main value of using this class is to accomplish the first step: decouple logic from interface.

    Parameters for the logic can be passed in the constructor or added via setters. Not in execute().
    This class makes no assumptions nor leads towards any of those strategies.
    The driving factor of this class is to provide a place where to extract the logic from some interface code.
    Browsing the subclasses of this class you can see all the logic in your application that has been isolated
    but not yet modeled into a proper Interaction.
    Thus, the list of subclasses of this class serves as a sort of To-Do list.
    """

    def execute(self):
        raise NotImplementedError('Subclass responsibility')
